import math
import numbers

import numpy as np
from sklearn.utils.validation import check_scalar

from ..utils.validation import _check_ts_array, check_array, check_type
from . import _cmatrix_profile
from ._distance import _format_return


def matrix_profile(  # noqa: PLR0912
    x,
    y=None,
    *,
    window=5,
    dim=0,
    exclude=None,
    n_jobs=-1,
    return_index=False,
):
    """
    Compute the matrix profile.

    - If only ``x`` is given, compute the similarity self-join of every subsequence in
      ``x`` of size ``window`` to its nearest neighbor in `x` excluding trivial matches
      according to the ``exclude`` parameter.
    - If both ``x`` and ``y`` are given, compute the similarity join of every
      subsequenec in ``y`` of size ``window`` to its nearest neighbor in ``x`` excluding
      matches according to the ``exclude`` parameter.

    Parameters
    ----------
    x : array-like of shape (n_timestep, ), (n_samples, xn_timestep) or \
        (n_samples, n_dim, xn_timestep)
        The first time series.
    y : array-like of shape (n_timestep, ), (n_samples, yn_timestep) or \
        (n_samples, n_dim, yn_timestep), optional
        The optional second time series. y is broadcast to the shape of x if possible.
    window : int or float, optional
        The subsequence size, by default 5

        - if float, a fraction of `y.shape[-1]`.
        - if int, the exact subsequence size.
    dim : int, optional
        The dim to compute the matrix profile for, by default 0.
    exclude : int or float, optional
        The size of the exclusion zone. The default exclusion zone is  0.2 for
        similarity self-join and 0.0 for similarity join.

        - if float, expressed as a fraction of the windows size.
        - if int, exact size (0 >= exclude < window).
    n_jobs : int, optional
        The number of jobs to use when computing the profile.
    return_index : bool, optional
        Return the matrix profile index.

    Returns
    -------
    mp : ndarray of shape (profile_size, ) or (n_samples, profile_size)
        The matrix profile.
    mpi : ndarray of shape (profile_size, ) or (n_samples, profile_size), optional
        The matrix profile index.

    Notes
    -----
    The `profile_size` depends on the input.

    - If `y` is `None`, `profile_size` is  ``x.shape[-1] - window + 1``
    - If `y` is not `None`, `profile_size` is ``y.shape[-1] - window + 1``

    References
    ----------
    Yeh, C. C. M. et al. (2016).
        Matrix profile I: All pairs similarity joins for time series: a unifying view
        that includes motifs, discords and shapelets. In 2016 IEEE 16th international
        conference on data mining (ICDM)
    """
    x = check_array(x, allow_3d=True, ensure_2d=False)

    if y is not None:
        y = check_array(y, allow_3d=True, ensure_2d=False)
        if x.ndim > 1:
            y = np.broadcast_to(y, x.shape)

        if x.ndim != y.ndim:
            raise ValueError("Both x and y must have the same dimensionality")
        if x.ndim != 1 and x.shape[0] != y.shape[0]:
            raise ValueError("Both x and y must have the same number of samples")
        if x.ndim > 2 and x.shape[1] != y.shape[1]:
            raise ValueError("Both x and y must have the same number of dimensions")
        if not y.shape[-1] <= x.shape[-1]:
            raise ValueError(
                "y.shape[-1] > x.shape[-1]. If you want to compute the matrix profile "
                "of the similarity join of YX, swap the order of inputs."
            )
        exclude = exclude if exclude is not None else 0.0
    else:
        y = x
        exclude = exclude if exclude is not None else 0.2

    if x.ndim > 2 and not 0 <= dim < x.shape[1]:
        raise ValueError("Invalid dim (%d)" % x.shape[1])

    check_type(window, "window", (numbers.Integral, numbers.Real))
    check_type(exclude, "exclude", (numbers.Integral, numbers.Real))
    if isinstance(window, numbers.Integral):
        check_scalar(
            window,
            "window",
            numbers.Integral,
            min_val=1,
            max_val=min(y.shape[-1], x.shape[-1]),
        )
    elif isinstance(window, numbers.Real):
        check_scalar(
            window,
            "window",
            numbers.Real,
            min_val=0,
            max_val=1,
            include_boundaries="right",
        )
        window = math.ceil(window * y.shape[-1])

    if isinstance(exclude, numbers.Integral):
        check_scalar(exclude, "exclude", numbers.Integral, min_val=1)
    elif isinstance(exclude, numbers.Real):
        check_scalar(
            exclude,
            "exclude",
            numbers.Real,
            min_val=0,
        )
        exclude = math.ceil(window * exclude)

    mp, mpi = _cmatrix_profile._paired_matrix_profile(
        _check_ts_array(x),
        _check_ts_array(y),
        window,
        dim,
        exclude,
        n_jobs,
    )

    if return_index:
        return _format_return(mp, 2, x.ndim), _format_return(mpi, 2, x.ndim)
    else:
        return _format_return(mp, 2, x.ndim)



import warnings

import numpy as np
from sklearn.base import clone
from sklearn.neighbors import KNeighborsClassifier as Sklearn_KNeighborsClassifier
from sklearn.utils.validation import check_is_fitted

from ...distance import KNeighborsClassifier, paired_distance
from ...ensemble import ExtraShapeletTreesClassifier, ShapeletForestClassifier
from ...utils.validation import check_array
from ._nn import KNeighborsCounterfactual
from ._proto import PrototypeCounterfactual
from ._sf import ShapeletForestCounterfactual


def _proximity(
    x_true,
    x_counterfactuals,
    metric="normalized_euclidean",
    metric_params=None,
):
    """
    Compute the proximity of the counterfactuals.

    Parameters
    ----------
    x_true : array-like of shape (n_samples, n_timestep)
        The true samples.
    x_counterfactuals : array-like of shape (n_samples, n_timestep)
        The counterfactual samples.
    metric : str or callable, optional
        The distance metric

        See ``_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_metrics>`.

    Returns
    -------
    ndarray
        The scores.
    """
    x_true = check_array(x_true, allow_3d=True, input_name="x_true")
    x_counterfactuals = check_array(
        x_counterfactuals, allow_3d=True, input_name="x_counterfactuals"
    )

    return paired_distance(
        x_true,
        x_counterfactuals,
        metric=metric,
        metric_params=metric_params,
        dim="mean",
    )


_COUNTERFACTUALS = {
    "prototype": PrototypeCounterfactual,
}


def _best_counterfactional(estimator):
    """
    Infer the counterfactual explainer to use based on the estimator.

    Parameters
    ----------
    estimator : object
        The estimator.

    Returns
    -------
    class
        The counterfactual explainer class.
    """
    if (
        isinstance(estimator, (ShapeletForestClassifier, ExtraShapeletTreesClassifier))
        and estimator.metric == "euclidean"
    ):
        return ShapeletForestCounterfactual
    elif isinstance(estimator, (Sklearn_KNeighborsClassifier, KNeighborsClassifier)):
        return KNeighborsCounterfactual
    else:
        return PrototypeCounterfactual


def counterfactuals(  # noqa: PLR0912
    estimator,
    x,
    y,
    *,
    train_x=None,
    train_y=None,
    method="best",
    scoring="deprecated",
    valid_scoring="deprecated",
    proximity=None,
    random_state=None,
    method_args=None,
):
    """
    Compute a single counterfactual example for each sample.

    Parameters
    ----------
    estimator : object
        The estimator used to compute the counterfactual example.
    x : array-like of shape (n_samples, n_timestep)
        The data samples to fit counterfactuals to.
    y : array-like broadcast to shape (n_samples,)
        The desired label of the counterfactual.
    train_x : array-like of shape (n_samples, n_timestep), optional
        Training samples if required by the explainer.
    train_y : array-like of shape (n_samples, ), optional
        Training labels if required by the explainer.
    method : str or BaseCounterfactual, optional
        The method to generate counterfactual explanations

        - if 'best', infer the most appropriate counterfactual explanation
          method based on the estimator.

        .. versionchanged:: 1.1.0

        - if str, select counterfactual explainer from named collection. See
          ``_COUNTERFACTUALS.keys()`` for a list of valid values.
        - if, BaseCounterfactual use the supplied counterfactual.
    scoring : str, callable, list or dict, optional
        The scoring function to determine the similarity between the counterfactual
        sample and the original sample.

        .. deprecated:: 1.1
            ``scoring`` was renamed to ``proximity`` in 1.1 and will be removed in 1.2.
    valid_scoring : bool, optional
        Only compute score for successful counterfactuals.

        .. deprecated:: 1.1
            ``valid_scoring`` will be removed in 1.2.
    proximity : str, callable, list or dict, optional
        The scoring function to determine the similarity between the counterfactual
        sample and the original sample.
    random_state : RandomState or int, optional
        The pseudo random number generator to ensure stable result.
    method_args : dict, optional
        Optional arguments to the counterfactual explainer.

        .. versionadded:: 1.1.0

    Returns
    -------
    x_counterfactuals : ndarray of shape (n_samples, n_timestep)
        The counterfactual example.
    valid : ndarray of shape (n_samples,)
        Indicator matrix for valid counterfactuals.
    score : ndarray of shape (n_samples,) or dict, optional
        Return score of the counterfactual transform, if ``scoring`` is not None.
    """
    check_is_fitted(estimator)
    if method_args is None:
        method_args = {}

    if isinstance(method, str):
        if method == "best":
            Explainer = _best_counterfactional(estimator)
        else:
            Explainer = _COUNTERFACTUALS.get(method)
            if Explainer is None:
                raise ValueError(
                    "method should be %s, got %r"
                    % (set(_COUNTERFACTUALS.keys()), method)
                )

        if "train_x" in method_args or "train_y" in method_args:
            train_x = method_args.pop("train_x")
            train_y = method_args.pot("train_y")
            warnings.warn(
                "train_x and train_y as method_args has been deprecated in 1.1 and "
                "will be removed in 1.2. Use the train_x and train_y keyword params.",
                FutureWarning,
            )

        explainer = Explainer(**method_args)
    else:
        explainer = clone(method)

    if random_state is not None and "random_state" in explainer.get_params():
        explainer.set_params(random_state=random_state)

    explainer.fit(estimator, train_x, train_y)
    y = np.broadcast_to(y, (x.shape[0],))
    x_counterfactuals = explainer.explain(x, y)
    valid = estimator.predict(x_counterfactuals) == y
    if scoring != "deprecated":
        warnings.warn(
            "'scoring' was renamed to 'proximity' 1.1 and will be removed in 1.2.",
            FutureWarning,
        )
        proximity = scoring

    if valid_scoring != "deprecated":
        warnings.warn(
            "'valid_scoring' has been deprecated in 1.1 and will be removed in 1.2",
            FutureWarning,
        )
    else:
        valid_scoring = False

    if valid_scoring is True:
        x = x[valid]
        x_counterfactuals = x_counterfactuals[valid]

    if proximity is not None:
        sc = _proximity(x, x_counterfactuals, metric=proximity)
        return x_counterfactuals, valid, sc
    else:
        return x_counterfactuals, valid



# Authors: Isak Samsten
# License: BSD 3 clause
"""Base classes for all estimators."""

import itertools
import warnings

import numpy as np
from sklearn.base import BaseEstimator as SklearnBaseEstimator
from sklearn.utils.fixes import parse_version
from sklearn.utils.validation import _check_y, check_is_fitted

from . import __version__
from .utils.validation import _num_timesteps, check_array, check_X_y

__all__ = [
    "BaseEstimator",
    "CounterfactualMixin",
    "ExplainerMixin",
    "is_counterfactual",
    "is_explainer",
]

_DEFAULT_TAGS = {
    # The estimator can be fit with variable length time series
    "allow_eos": False,
    # By default no estimator supports nan.
    "allow_nan": False,
    # The explainer requires an estimator to fit
    "requires_estimator": False,
    # X_types = ["3darray"]
}


class BaseEstimator(SklearnBaseEstimator):
    """Base estimator for all Wildboar estimators."""

    _doc_link_module = "wildboar"

    @property
    def _doc_link_template(self):
        wildboar_version = parse_version(__version__)
        if wildboar_version.dev is None:
            version_url = f"{wildboar_version.major}.{wildboar_version.minor}"
        else:
            version_url = "master"
        return (
            "https://wildboar.dev/%s/api/{module_path}/index.html#{estimator_module}.{estimator_name}"
            % version_url
        )

    def _doc_link_url_param_generator(self, other):
        estimator_name = self.__class__.__name__
        estimator_module = ".".join(
            itertools.takewhile(
                lambda part: not part.startswith("_"),
                self.__class__.__module__.split("."),
            )
        )
        return {
            "module_path": estimator_module.replace(".", "/"),
            "estimator_module": estimator_module,
            "estimator_name": estimator_name,
        }

    # Same additions as scikit-learn
    def __getstate__(self):
        """Get the state of the estimator.

        Add a new element to the dict we return `_wildboar_version` which
        is we use to warn when setting the state.

        Returns
        -------
        dict
            The state
        """
        try:
            state = super().__getstate__()
        except AttributeError:
            state = self.__dict__.copy()

        if type(self).__module__.startswith("wildboar."):
            return dict(state.items(), _wildboar_version=__version__)
        else:
            return state

    # Same check as scikit-learn
    def __setstate__(self, state):
        """Set the state of the estimator.

        Gives a warning if a user tries to unpickle a object serialized
        from a older version of Wildboar.

        Parameters
        ----------
        state : The state
        """
        if type(self).__module__.startswith("wildboar."):
            pickle_version = state.pop("_wildboar_version", "pre-1.1")
            if pickle_version != __version__:
                warnings.warn(
                    "Trying to unpickle estimator {0} from version {1} when "
                    "using version {2}. This might lead to breaking code or "
                    "invalid results. Use at your own risk.".format(
                        self.__class__.__name__, pickle_version, __version__
                    ),
                    UserWarning,
                )
        try:
            super().__setstate__(state)
        except AttributeError:
            self.__dict__.update(state)

    def _more_tags(self):
        return _DEFAULT_TAGS

    # Disable feature names since there are no feature names
    # NOTE: Consider adding support for named dimensions for multivariate time series.
    def _check_feature_names(self, X, *, reset):
        pass

    def _check_n_features(self, X, reset):
        self._check_n_timesteps(X, reset)

    def _check_n_timesteps(self, X, reset):
        try:
            n_timesteps, n_dims = _num_timesteps(X)
        except TypeError as e:
            if not reset and hasattr(self, "n_timesteps_in_"):
                raise ValueError(
                    "X does not contain any timesteps, but "
                    f"{self.__class__.__name__} is expecting "
                    f"{self.n_timesteps_in_} timesteps"
                ) from e
            # If the number of timesteps is not defined and reset=True,
            # then we skip this check
            return

        if reset:
            self.n_timesteps_in_ = n_timesteps
            self.n_dims_in_ = n_dims

            # Set n_features_in_ for compatibility with scikit-learn
            self.n_features_in_ = n_timesteps * n_dims
            return

        if not hasattr(self, "n_timesteps_in_") and not hasattr(self, "n_dims_in_"):
            # Skip this check if the expected number of expected input features
            # was not recorded by calling fit first. This is typically the case
            # for stateless transformers.
            return

        if n_timesteps != self.n_timesteps_in_:
            raise ValueError(
                f"X has {n_timesteps} timesteps, but {self.__class__.__name__} "
                f"is expecting {self.n_timesteps_in_} timesteps as input."
            )
        if n_dims != self.n_dims_in_:
            raise ValueError(
                f"X has {n_dims} dimensions, but {self.__class__.__name__} "
                f"is expecting {self.n_dims_in_} dimensions as input."
            )

    def _validate_force_n_dims(self, X):
        """Validate that X is valid.

        The validation takes into consideration the optional _force_n_dims
        attribute.

        Parameters
        ----------
        X : array-like of shape (n_samples, m)
            The array to validate.

        Returns
        -------
        X : ndarray of shape (n_samples, self._force_n_dims, -1)
            The array validated and reshaped
        """
        if hasattr(self, "_force_n_dims"):
            X = np.asarray(X)
            if X.ndim == 2:
                return X.reshape(X.shape[0], self._force_n_dims, -1)
            elif X.ndim == 3 and X.shape[1] != self._force_n_dims:
                raise ValueError(
                    f"{self.__class__.__name__} has _force_n_dims set to "
                    f"{self._force_n_dims}, but X has {X.shape[1]} dimensions as input."
                )

        return X

    # We override sklearn but delegate to wildboar's check_array
    # and check_X_y
    def _validate_data(
        self,
        X="no_validation",
        y="no_validation",
        reset=True,
        validate_separately=False,
        **check_params,
    ):
        if y is None and self._get_tags()["requires_y"]:
            raise ValueError(
                f"This {self.__class__.__name__} estimator "
                "requires y to be passed, but the target y is None."
            )

        no_val_X = isinstance(X, str) and X == "no_validation"
        no_val_y = y is None or isinstance(y, str) and y == "no_validation"

        default_check_params = {"estimator": self}
        check_params = {**default_check_params, **check_params}

        if not no_val_X:
            X = self._validate_force_n_dims(X)

        if no_val_X and no_val_y:
            raise ValueError("Validation should be done on X, y or both.")
        elif not no_val_X and no_val_y:
            X = check_array(X, input_name="X", **check_params)
            out = X
        elif no_val_X and not no_val_y:
            y = _check_y(y, **check_params)
            out = y
        else:
            if validate_separately:
                # We need this because some estimators validate X and y
                # separately, and in general, separately calling check_array()
                # on X and y isn't equivalent to just calling check_X_y()
                # :(
                check_X_params, check_y_params = validate_separately
                if "estimator" not in check_X_params:
                    check_X_params = {**default_check_params, **check_X_params}
                X = check_array(X, input_name="X", **check_X_params)
                if "estimator" not in check_y_params:
                    check_y_params = {**default_check_params, **check_y_params}
                y = check_array(y, input_name="y", **check_y_params)
            else:
                X, y = check_X_y(X, y, **check_params)
            out = X, y

        if not no_val_X:
            self._check_n_timesteps(X, reset=reset)

        return out


class ExplainerMixin:
    """Mixin class for all explainers in wildboar."""

    _estimator_type = "explainer"

    def _validate_estimator(self, estimator, allow_3d=False):
        """Check the estimator.

        Set the `n_timesteps_in_` and `n_dims_in_` parameter from the estimator.

        Parameters
        ----------
        estimator : object
            The estimator object to check

        allow_3d : bool, optional
            If estimator fit with 3d-arrays are supported.
        """
        check_is_fitted(estimator)
        if hasattr(estimator, "n_timesteps_in_"):
            self.n_timesteps_in_ = estimator.n_timesteps_in_
        elif hasattr(estimator, "n_features_in_"):
            self.n_timesteps_in_ = estimator.n_features_in_
        else:
            raise ValueError(
                "Unable to find the number of timesteps from {}. Please ensure that "
                "the estimator is correctly fit and sets the n_timesteps_in_ or "
                "n_features_in_ properties.".format(type(estimator).__qualname__)
            )

        self.n_features_in_ = self.n_timesteps_in_
        if hasattr(estimator, "n_dims_in_"):
            self.n_dims_in_ = estimator.n_dims_in_
        else:
            self.n_dims_in_ = 1

        if self.n_dims_in_ > 1 and not allow_3d:
            raise ValueError(
                "The explainer does not permit 3darrays as input but the estimator is "
                "fitted with a 3darray of shape (?, {}, {}).".format(
                    self.n_dims_in_, self.n_timesteps_in_
                )
            )

        return estimator

    def fit_explain(self, estimator, x=None, y=None, **kwargs):
        """Fit and return the explanation.

        Parameters
        ----------
        estimator : Estimator
            The estimator to explain.
        x : time-series, optional
            The input time series.
        y : array-like of shape (n_samples, ), optional
            The labels.
        **kwargs
            Optional extra arguments.

        Returns
        -------
        ndarray
            The explanation.
        """
        return self.fit(estimator, x, y, **kwargs).explain(x, y)

    def plot(self, x=None, y=None, ax=None):
        """Plot the explanation.

        Returns
        -------
        ax : Axes
            The axes object
        """
        from .utils.plot import plot_time_domain

        plot_time_domain(self.explain(x, y), y, ax=ax)

    def _more_tags(self):
        return {"requires_estimator": True}


class CounterfactualMixin:
    """Mixin class for counterfactual explainer."""

    _estimator_type = "counterfactual"

    def score(self, x, y):
        """Score the counterfactual explainer in terms of closeness of fit.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timestep)
            The samples.

        y : array-like of shape (n_samples, )
            The desired counterfactal label.

        Returns
        -------
        score : float
            The closensess of fit.
        """
        from .metrics import proximity_score

        return proximity_score(x, self.explain(x, y))


def is_counterfactual(estimator):
    """Check if estimator is a counterfactual explainer.

    Parameters
    ----------
    estimator : object
        The estimator

    Returns
    -------
    bool
        True if the estimator probably is a counterfactual explainer
    """
    return (
        hasattr(estimator, "explain")
        and hasattr(type(estimator), "_estimator_type")
        and type(estimator)._estimator_type == "counterfactual"
    )


def is_explainer(estimator):
    """Check if estimator is an explainer.

    Parameters
    ----------
    estimator : object
        The estimator

    Returns
    -------
    bool
        True if the estimator probably is an explainer
    """
    return hasattr(estimator, "explain")



# cython: language_level=3

# Authors: Isak Samsten
# License: BSD 3 clause

cdef struct HeapElement:
    double value
    Py_ssize_t index

cdef class Heap:

    cdef HeapElement* heap
    cdef Py_ssize_t n_elements
    cdef Py_ssize_t max_elements

    cdef void push(self, Py_ssize_t index, double value) noexcept nogil

    cdef HeapElement maxelement(self) noexcept nogil

    cdef double maxvalue(self) noexcept nogil

    cdef bint isempty(self) noexcept nogil

    cdef bint isfull(self) noexcept nogil

    cdef void reset(self) noexcept nogil

    cdef HeapElement getelement(self, Py_ssize_t i) noexcept nogil

cdef class List:
    cdef list py_list
    cdef readonly Py_ssize_t size

    cdef void* get(self, Py_ssize_t i) noexcept nogil

ctypedef fused double_or_int:
    Py_ssize_t
    double

cdef void argsort(
    double_or_int *values, Py_ssize_t *order, Py_ssize_t length
) noexcept nogil

cdef int realloc_array(
    void** a, Py_ssize_t p, Py_ssize_t size, Py_ssize_t *cap
) except -1 nogil

cdef int safe_realloc(void** ptr, Py_ssize_t new_size) except -1 nogil

cdef object to_ndarray_int(Py_ssize_t *arr, Py_ssize_t n)

cdef object to_ndarray_double(double *arr, Py_ssize_t n)



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np

from libc.math cimport INFINITY, NAN, ceil, exp, fabs, log2
from libc.stdlib cimport calloc, free, malloc
from libc.string cimport memcpy, memset
from numpy cimport uint32_t

from scipy.sparse import csr_matrix

from ..distance._cdistance cimport Metric

from ..transform._attr_gen cimport Attribute, AttributeGenerator
from ..utils cimport TSArray
from ..utils._misc cimport List, argsort, safe_realloc
from ..utils._rand cimport RAND_R_MAX, rand_int, rand_uniform


cdef double ATTRIBUTE_THRESHOLD = 1e-7


cdef struct SplitPoint:
    Py_ssize_t split_point
    double threshold
    double impurity_improvement
    double impurity_left
    double impurity_right
    Attribute attribute


cdef class TreeAttributeGenerator:
    cdef AttributeGenerator generator

    def __init__(self, AttributeGenerator generator):
        self.generator = generator

    def __reduce__(self):
        return self.__class__, (self.generator, )

    cdef int reset(self, TSArray X) noexcept nogil:
        return self.generator.reset(X)

    cdef Py_ssize_t get_n_attributess(self, TSArray X, Py_ssize_t depth) noexcept nogil:
        return self.generator.get_n_attributess(X)

    cdef Py_ssize_t next_attribute(
        self,
        Py_ssize_t attribute_id,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        Attribute *transient,
        uint32_t *seed
    ) noexcept nogil:
        return self.generator.next_attribute(
            attribute_id, X, samples, n_samples, transient, seed
        )

    cdef Py_ssize_t init_persistent(
        self,
        TSArray X,
        Attribute *transient,
        Attribute *persistent
    ) noexcept nogil:
        return self.generator.init_persistent(X, transient, persistent)

    cdef Py_ssize_t free_transient(self, Attribute *attribute) noexcept nogil:
        return self.generator.free_transient(attribute)

    cdef Py_ssize_t free_persistent(self, Attribute *attribute) noexcept nogil:
        return self.generator.free_persistent(attribute)

    cdef double transient_value(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample
    ) noexcept nogil:
        return self.generator.transient_value(attribute, X, sample)

    cdef double persistent_value(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample
    ) noexcept nogil:
        return self.generator.persistent_value(attribute, X, sample)

    cdef void transient_values(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double* values
    ) noexcept nogil:
        self.generator.transient_values(
            attribute, X, samples, n_samples, values
        )

    cdef void persistent_values(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double* values
    ) noexcept nogil:
        self.generator.persistent_values(
            attribute, X, samples, n_samples, values
        )

    cdef object persistent_to_object(self, Attribute *attribute):
        return self.generator.persistent_to_object(attribute)

    cdef Py_ssize_t persistent_from_object(self, object object, Attribute *attribute):
        return self.generator.persistent_from_object(object, attribute)


cdef class DynamicTreeAttributeGenerator(TreeAttributeGenerator):

    cdef double alpha

    def __init__(self, AttributeGenerator generator, double alpha):
        super().__init__(generator)
        self.alpha = alpha

    def __reduce__(self):
        return self.__class__, (self.generator, self.alpha)

    cdef Py_ssize_t get_n_attributess(self, TSArray X, Py_ssize_t depth) noexcept nogil:
        cdef Py_ssize_t n_attributes = self.generator.get_n_attributess(X)
        cdef double weight = 1.0 - exp(-fabs(self.alpha) * depth)
        if self.alpha < 0:
            weight = 1 - weight

        return <Py_ssize_t> max(1, ceil(n_attributes * weight))


cdef class Criterion:

    cdef double weighted_n_left
    cdef double weighted_n_right
    cdef double weighted_n_total
    cdef Py_ssize_t start
    cdef Py_ssize_t end
    cdef Py_ssize_t *samples
    cdef const double[:] sample_weight

    cdef void init(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t *samples,
        const double[:] sample_weights,
    ) noexcept nogil:
        self.start = start
        self.end = end
        self.samples = samples
        self.sample_weight = sample_weights

    cdef void reset(self) noexcept nogil:
        pass

    cdef void update(self, Py_ssize_t pos, Py_ssize_t new_pos) noexcept nogil:
        pass

    cdef double proxy_impurity(self) noexcept nogil:
        cdef double left_impurity
        cdef double right_impurity
        self.child_impurity(&left_impurity, &right_impurity)
        return -self.weighted_n_right * right_impurity - self.weighted_n_left * left_impurity

    cdef double impurity(self) noexcept nogil:
        pass

    cdef void child_impurity(self, double *left, double *right) noexcept nogil:
        pass

    cdef double impurity_improvement(
        self,
        double impurity_parent,
        double impurity_left,
        double impurity_right,
        double weighted_n_samples,
    ) noexcept nogil:
        return ((self.weighted_n_total / weighted_n_samples) *
                (impurity_parent - (self.weighted_n_right / self.weighted_n_total * impurity_right)
                                 - (self.weighted_n_left / self.weighted_n_total * impurity_left)))

    cdef void leaf_value(self, Tree tree, Py_ssize_t node_id) noexcept nogil:
        pass

cdef class ClassificationCriterion(Criterion):

    cdef const Py_ssize_t[:] labels
    cdef Py_ssize_t n_labels
    cdef double *sum_left
    cdef double *sum_right
    cdef double *sum_total

    def __cinit__(self, const Py_ssize_t[:] y, Py_ssize_t n_labels):
        self.labels = y
        self.n_labels = n_labels
        self.sum_left = <double*> calloc(n_labels, sizeof(double))
        self.sum_right = <double*> calloc(n_labels, sizeof(double))
        self.sum_total = <double*> calloc(n_labels, sizeof(double))

    def __dealloc__(self):
        free(self.sum_total)
        free(self.sum_left)
        free(self.sum_right)

    cdef void init(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t *samples,
        double[:] sample_weights,
    ) noexcept nogil:
        Criterion.init(self, start, end, samples, sample_weights)
        self.weighted_n_total = 0

        memset(self.sum_total, 0, self.n_labels * sizeof(double))

        cdef Py_ssize_t i, j
        cdef double w = 1.0
        for i in range(start, end):
            j = samples[i]
            if sample_weights is not None:
                w = sample_weights[j]

            self.sum_total[self.labels[j]] += w
            self.weighted_n_total += w

        self.reset()

    cdef void reset(self) noexcept nogil:
        self.weighted_n_left = 0
        self.weighted_n_right = self.weighted_n_total
        memset(self.sum_left, 0, self.n_labels * sizeof(double))
        memcpy(self.sum_right, self.sum_total, self.n_labels * sizeof(double))

    cdef void update(self, Py_ssize_t pos, Py_ssize_t new_pos) noexcept nogil:
        cdef Py_ssize_t i, j
        cdef double w = 1.0

        for i in range(pos, new_pos):
            j = self.samples[i]
            if self.sample_weight is not None:
                w = self.sample_weight[j]

            self.sum_left[self.labels[j]] += w
            self.weighted_n_left += w

        self.weighted_n_right = self.weighted_n_total - self.weighted_n_left
        for i in range(self.n_labels):
            self.sum_right[i] = self.sum_total[i] - self.sum_left[i]

    cdef double impurity(self) noexcept nogil:
        pass

    cdef void child_impurity(self, double* left, double *right) noexcept nogil:
        pass

    cdef void leaf_value(self, Tree tree, Py_ssize_t node_id) noexcept nogil:
        cdef Py_ssize_t i
        cdef double prob
        for i in range(self.n_labels):
            prob = self.sum_total[i] / self.weighted_n_total
            tree.set_leaf_value(node_id, i, prob)

cdef class GiniCriterion(ClassificationCriterion):

    cdef double impurity(self) noexcept nogil:
        cdef double sq_count = 0.0
        cdef double c
        cdef Py_ssize_t i
        for i in range(self.n_labels):
            c = self.sum_total[i]
            sq_count += c * c

        return 1.0 - sq_count / (self.weighted_n_total * self.weighted_n_total)

    cdef void child_impurity(self, double *left, double *right) noexcept nogil:
        cdef double sq_left = 0
        cdef double sq_right = 0
        cdef double v
        cdef Py_ssize_t i

        for i in range(self.n_labels):
            v = self.sum_left[i]
            sq_left += v * v

            v = self.sum_right[i]
            sq_right += v * v

        left[0] = 1 - sq_left / (self.weighted_n_left * self.weighted_n_left)
        right[0] = 1 - sq_right / (self.weighted_n_right * self.weighted_n_right)

cdef class EntropyCriterion(ClassificationCriterion):

    cdef double impurity(self) noexcept nogil:
        cdef double c
        cdef double entropy = 0
        cdef Py_ssize_t i
        for i in range(self.n_labels):
            c = self.sum_total[i]
            if c > 0:
                c /= self.weighted_n_total
                entropy -= c * log2(c)

        return entropy

    cdef void child_impurity(self, double *left, double *right) noexcept nogil:
        left[0] = 0
        right[0] = 0
        cdef double v
        cdef Py_ssize_t i

        for i in range(self.n_labels):
            v = self.sum_left[i]
            if v > 0:
                v /= self.weighted_n_left
                left[0] -= v * log2(v)

            v = self.sum_right[i]
            if v > 0:
                v /= self.weighted_n_right
                right[0] -= v * log2(v)

cdef class RegressionCriterion(Criterion):
    cdef double sum_left
    cdef double sum_right
    cdef double sum_total
    cdef double sum_sq_total
    cdef const double[:] labels
    cdef Py_ssize_t pos

    def __cinit__(self, const double[:] y):
        self.labels = y
        self.sum_left = 0
        self.sum_right = 0
        self.sum_total = 0
        self.sum_sq_total = 0

    cdef void init(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t *samples,
        double[:] sample_weights,
    ) noexcept nogil:
        Criterion.init(self, start, end, samples, sample_weights)
        self.sum_total = 0
        self.sum_sq_total = 0
        self.weighted_n_total = 0

        cdef Py_ssize_t i, j
        cdef double x
        cdef double w = 1.0

        for i in range(start, end):
            j = samples[i]
            if sample_weights is not None:
                w = sample_weights[j]

            x = w * self.labels[j]
            self.sum_total += x
            self.sum_sq_total += x * x
            self.weighted_n_total += w

        self.reset()
        self.start = start

    cdef void reset(self) noexcept nogil:
        self.weighted_n_left = 0
        self.weighted_n_right = self.weighted_n_total
        self.sum_left = 0
        self.sum_right = self.sum_total
        self.pos = 0

    cdef void update(self, Py_ssize_t pos, Py_ssize_t new_pos) noexcept nogil:
        cdef Py_ssize_t i
        cdef Py_ssize_t j
        cdef double w = 1.0
        for i in range(pos, new_pos):
            j = self.samples[i]
            if self.sample_weight is not None:
                w = self.sample_weight[j]

            self.sum_left += w * self.labels[j]
            self.weighted_n_left += w

        self.weighted_n_right = self.weighted_n_total - self.weighted_n_left
        self.sum_right = self.sum_total - self.sum_left
        self.pos = new_pos

    cdef void leaf_value(self, Tree tree, Py_ssize_t node_id) noexcept nogil:
        tree.set_leaf_value(node_id, 0, self.sum_total / self.weighted_n_total)

cdef class MSECriterion(RegressionCriterion):

    cdef double proxy_impurity(self) noexcept nogil:
        cdef double proxy_impurity_left = self.sum_left * self.sum_left
        cdef double proxy_impurity_right = self.sum_right * self.sum_right
        return proxy_impurity_left / self.weighted_n_left + proxy_impurity_right / self.weighted_n_right

    cdef double impurity(self) noexcept nogil:
        cdef double impurity
        impurity = self.sum_sq_total / self.weighted_n_total
        impurity -= (self.sum_total / self.weighted_n_total) ** 2
        return impurity

    cdef void child_impurity(self, double* left, double *right) noexcept nogil:
        cdef double left_sq_sum = 0
        cdef double right_sq_sum = 0
        cdef double w = 1.0
        cdef double y
        cdef Py_ssize_t i, j

        for i in range(self.start, self.pos):
            j = self.samples[i]
            if self.sample_weight is not None:
                w = self.sample_weight[j]

            y = self.labels[j]
            left_sq_sum += w * y * y
        right_sq_sum = self.sum_sq_total - left_sq_sum

        left[0] = left_sq_sum / self.weighted_n_left
        left[0] -= (self.sum_left / self.weighted_n_left) ** 2
        right[0] = right_sq_sum / self.weighted_n_right
        right[0] -= (self.sum_right / self.weighted_n_right) ** 2

cpdef Tree _make_tree(
    TreeAttributeGenerator generator,
    Py_ssize_t n_labels,
    Py_ssize_t max_depth,
    list attributes,
    object threshold,
    object value,
    object left,
    object right,
    object impurity,
    object n_node_samples,
    object n_weighted_node_samples
):
    cdef Tree tree = Tree(generator, n_labels, capacity=len(attributes) + 1)
    tree._max_depth = max_depth
    cdef Py_ssize_t node_count = len(attributes)
    cdef Py_ssize_t i
    cdef Py_ssize_t dim
    cdef object arr
    cdef Attribute *attribute
    cdef object value_reshape = value.reshape(-1)

    tree._node_count = node_count
    for i in range(node_count):
        if attributes[i] is not None:
            attribute = <Attribute*> malloc(sizeof(Attribute))
            generator.persistent_from_object(attributes[i], attribute)
            tree._attributes[i] = attribute
        else:
            tree._attributes[i] = NULL
        tree._thresholds[i] = threshold[i]
        tree._left[i] = left[i]
        tree._right[i] = right[i]
        tree._impurity[i] = impurity[i]
        tree._n_node_samples[i] = n_node_samples[i]
        tree._n_weighted_node_samples[i] = n_weighted_node_samples[i]

    for i in range(node_count * n_labels):
        tree._values[i] = value_reshape[i]
    return tree


cdef class Tree:

    cdef TreeAttributeGenerator generator

    cdef Py_ssize_t _max_depth
    cdef Py_ssize_t _capacity
    cdef Py_ssize_t _n_labels  # 1 for regression

    cdef Py_ssize_t _node_count
    cdef Py_ssize_t *_left
    cdef Py_ssize_t *_right
    cdef Attribute **_attributes
    cdef double *_thresholds
    cdef double *_impurity
    cdef double *_values
    cdef double *_n_weighted_node_samples
    cdef Py_ssize_t *_n_node_samples

    def __cinit__(
        self,
        TreeAttributeGenerator generator,
        Py_ssize_t n_labels,
        Py_ssize_t capacity=10
    ):
        self.generator = generator
        self._node_count = 0
        self._capacity = capacity
        self._n_labels = n_labels
        self._attributes = <Attribute**> malloc(self._capacity * sizeof(Attribute*))
        self._thresholds = <double*> malloc(self._capacity * sizeof(double))
        self._values = <double*> malloc(self._capacity * self._n_labels * sizeof(double))
        self._left = <Py_ssize_t*> malloc(self._capacity * sizeof(Py_ssize_t))
        self._right = <Py_ssize_t*> malloc(self._capacity * sizeof(Py_ssize_t))
        self._impurity = <double*> malloc(self._capacity * sizeof(double))
        self._n_node_samples = <Py_ssize_t*> malloc(self._capacity * sizeof(Py_ssize_t))
        self._n_weighted_node_samples = <double*> malloc(self._capacity * sizeof(double))

    def __dealloc__(self):
        cdef Py_ssize_t i
        if self._attributes != NULL:
            for i in range(self._node_count):
                if self._attributes[i] != NULL:
                    self.generator.free_persistent(self._attributes[i])
                    free(self._attributes[i])
            free(self._attributes)

        if self._thresholds != NULL:
            free(self._thresholds)

        if self._values != NULL:
            free(self._values)

        if self._left != NULL:
            free(self._left)

        if self._right != NULL:
            free(self._right)

        if self._impurity != NULL:
            free(self._impurity)

        if self._n_node_samples != NULL:
            free(self._n_node_samples)

        if self._n_weighted_node_samples != NULL:
            free(self._n_weighted_node_samples)

    def __reduce__(self):
        return _make_tree, (
            self.generator,
            self._n_labels,
            self._max_depth,
            self.attribute,
            self.threshold,
            self.value,
            self.left,
            self.right,
            self.impurity,
            self.n_node_samples,
            self.n_weighted_node_samples,
        )

    @property
    def max_depth(self):
        return self._max_depth

    @property
    def value(self):
        cdef object arr = np.empty(self._node_count * self._n_labels, dtype=float)
        cdef Py_ssize_t i
        for i in range(self._n_labels * self._node_count):
            arr[i] = self._values[i]
        return arr.reshape(self._node_count, self._n_labels)

    @property
    def attribute(self):
        cdef Py_ssize_t i, j
        cdef Attribute* attribute
        cdef object object
        cdef list ret = []
        for i in range(self._node_count):
            attribute = self._attributes[i]
            if attribute != NULL:
                object = self.generator.persistent_to_object(attribute)
                ret.append(object)
            else:
                ret.append(None)
        return ret

    @property
    def feature(self):
        import warnings
        warnings.warn(
            "`feature` has been renamed to `attribute` in 1.2 and will be removed in 1.4",
            DeprecationWarning,
        )
        return self.attribute

    @property
    def n_node_samples(self):
        cdef object arr = np.zeros(self._node_count, dtype=np.intp)
        cdef Py_ssize_t i
        for i in range(self._node_count):
            arr[i] = self._n_node_samples[i]
        return arr

    @property
    def n_weighted_node_samples(self):
        cdef object arr = np.zeros(self._node_count, dtype=float)
        cdef Py_ssize_t i
        for i in range(self._node_count):
            arr[i] = self._n_weighted_node_samples[i]
        return arr

    @property
    def left(self):
        cdef object arr = np.empty(self._node_count, dtype=np.intp)
        cdef Py_ssize_t i
        for i in range(self._node_count):
            arr[i] = self._left[i]
        return arr

    @property
    def right(self):
        cdef object arr = np.empty(self._node_count, dtype=np.intp)
        cdef Py_ssize_t i
        for i in range(self._node_count):
            arr[i] = self._right[i]
        return arr

    @property
    def threshold(self):
        cdef object arr = np.empty(self._node_count, dtype=float)
        cdef Py_ssize_t i
        for i in range(self._node_count):
            arr[i] = self._thresholds[i]
        return arr

    @property
    def impurity(self):
        cdef object arr = np.empty(self._node_count, dtype=float)
        cdef Py_ssize_t i
        for i in range(self._node_count):
            arr[i] = self._impurity[i]
        return arr

    def predict(self, object X):
        cdef object apply = self.apply(X)
        return np.take(self.value, apply, axis=0, mode="clip")

    def apply(self, TSArray X):
        cdef Py_ssize_t[:] out = np.zeros((X.shape[0],), dtype=np.intp)
        cdef Attribute *attribute
        cdef double threshold, attribute_value
        cdef Py_ssize_t node_index
        cdef Py_ssize_t i
        with nogil:
            self.generator.reset(X)
            for i in range(X.shape[0]):
                node_index = 0
                while self._left[node_index] != -1:
                    threshold = self._thresholds[node_index]
                    attribute = self._attributes[node_index]
                    attribute_value = self.generator.persistent_value(
                        attribute, X, i
                    )
                    if attribute_value <= threshold:
                        node_index = self._left[node_index]
                    else:
                        node_index = self._right[node_index]

                out[i] = <Py_ssize_t> node_index

        return out.base

    def decision_path(self, TSArray X):
        cdef Py_ssize_t[:, :] out = np.zeros((X.shape[0], self.node_count), dtype=np.intp)
        cdef Py_ssize_t node_index
        cdef Py_ssize_t i
        cdef Attribute *attribute
        cdef double threshold, attribute_value
        with nogil:
            self.generator.reset(X)
            for i in range(X.shape[0]):
                node_index = 0
                while self._left[node_index] != -1:
                    out[i, node_index] = 1
                    threshold = self._thresholds[node_index]
                    attribute = self._attributes[node_index]
                    attribute_value = self.generator.persistent_value(
                        attribute, X, i
                    )
                    if attribute_value <= threshold:
                        node_index = self._left[node_index]
                    else:
                        node_index = self._right[node_index]

        return csr_matrix(out.base, dtype=bool)

    @property
    def node_count(self):
        return self._node_count

    cdef Py_ssize_t add_leaf_node(
        self,
        Py_ssize_t parent,
        bint is_left,
        Py_ssize_t n_node_samples,
        double n_weighted_node_samples,
    ) noexcept nogil:
        cdef Py_ssize_t node_id = self._node_count
        if node_id >= self._capacity:
            if self._increase_capacity() == -1:
                return -1

        self._n_node_samples[node_id] = n_node_samples
        self._n_weighted_node_samples[node_id] = n_weighted_node_samples
        if parent != -1:
            if is_left:
                self._left[parent] = node_id
            else:
                self._right[parent] = node_id
        self._left[node_id] = -1
        self._right[node_id] = -1
        self._impurity[node_id] = -1
        self._attributes[node_id] = NULL
        self._node_count += 1
        return node_id

    cdef void set_leaf_value(
        self,
        Py_ssize_t node_id,
        Py_ssize_t out_label,
        double out_value,
    ) noexcept nogil:
        self._values[out_label + node_id * self._n_labels] = out_value

    cdef Py_ssize_t add_branch_node(
        self,
        Py_ssize_t parent,
        bint is_left,
        Py_ssize_t n_node_samples,
        double n_weighted_node_samples,
        Attribute *attribute,
        double threshold,
        double impurity,
    ) noexcept nogil:
        cdef Py_ssize_t node_id = self._node_count
        if node_id >= self._capacity:
            if self._increase_capacity() == -1:
                return -1

        self._impurity[node_id] = impurity
        self._n_node_samples[node_id] = n_node_samples
        self._n_weighted_node_samples[node_id] = n_weighted_node_samples
        self._thresholds[node_id] = threshold
        self._attributes[node_id] = attribute
        if parent != -1:
            if is_left:
                self._left[parent] = node_id
            else:
                self._right[parent] = node_id

        self._node_count += 1
        return node_id

    cdef Py_ssize_t _increase_capacity(self) noexcept nogil:
        cdef Py_ssize_t new_capacity = self._node_count * 2
        cdef Py_ssize_t ret
        ret = safe_realloc(<void**> &self._attributes, sizeof(Attribute*) * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._thresholds, sizeof(double) * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._impurity, sizeof(double) * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._n_node_samples, sizeof(double) * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._n_weighted_node_samples, sizeof(Py_ssize_t) * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._values, sizeof(double) * self._n_labels * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._left, sizeof(Py_ssize_t) * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._right, sizeof(Py_ssize_t) * new_capacity)
        if ret == -1:
            return -1
        self._capacity = new_capacity
        return 0

cdef class TreeBuilder:

    # hyper-parameters
    cdef Py_ssize_t max_depth
    cdef Py_ssize_t min_sample_split
    cdef Py_ssize_t min_sample_leaf
    cdef double min_impurity_decrease

    # the id (in Tree) of the current node
    cdef Py_ssize_t current_node_id

    # the weight of the j:th sample
    cdef const double[:] sample_weights

    # the dataset of time series
    cdef TSArray X

    # the number of samples with non-zero weight
    cdef Py_ssize_t n_samples
    cdef double n_weighted_samples

    # buffer of samples from 0, ..., n_samples
    cdef Py_ssize_t *samples

    # temporary buffer of samples from 0, ..., n_samples
    cdef Py_ssize_t *samples_buffer

    # temporary buffer for attribute computations
    cdef double *attribute_buffer

    cdef TreeAttributeGenerator generator
    cdef Criterion criterion
    cdef Tree tree
    cdef uint32_t random_seed

    def __cinit__(
        self,
        TSArray X,
        const double[:] sample_weights,
        TreeAttributeGenerator generator,
        Criterion criterion,
        Tree tree,
        object random_state,
        Py_ssize_t max_depth=2**16,
        Py_ssize_t min_sample_split=2,
        Py_ssize_t min_sample_leaf=1,
        double min_impurity_decrease=0.0,
    ):
        self.max_depth = max_depth
        self.min_sample_split = min_sample_split
        self.min_sample_leaf = min_sample_leaf
        self.min_impurity_decrease = min_impurity_decrease
        self.random_seed = random_state.randint(0, RAND_R_MAX)

        self.X = X
        self.generator = generator
        self.criterion = criterion
        self.tree = tree

        self.current_node_id = 0
        self.n_samples = self.X.shape[0]
        self.samples = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * self.n_samples)
        self.samples_buffer = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * self.n_samples)
        self.attribute_buffer = <double*> malloc(sizeof(double) * self.n_samples)

        if (
            self.samples == NULL or
            self.attribute_buffer == NULL or
            self.samples_buffer == NULL
        ):
            raise MemoryError()

        cdef Py_ssize_t i
        cdef Py_ssize_t j = 0
        self.n_weighted_samples = 0.0
        for i in range(self.n_samples):
            if sample_weights is None or sample_weights[i] != 0.0:
                self.samples[j] = i
                j += 1
                if sample_weights is not None:
                    self.n_weighted_samples += sample_weights[i]
                else:
                    self.n_weighted_samples += 1.0

        self.n_samples = j
        self.sample_weights = sample_weights

    def __dealloc__(self):
        free(self.samples)
        free(self.samples_buffer)
        free(self.attribute_buffer)

    @property
    def tree_(self):
        return self.tree

    cpdef Py_ssize_t build_tree(self):
        cdef Py_ssize_t root_node_id
        cdef Py_ssize_t max_depth = 0
        with nogil:
            self.generator.reset(self.X)
            root_node_id = self._build_tree(
                0,
                self.n_samples,
                0,
                -1,
                False,
                NAN,
                &max_depth
            )
        self.tree._max_depth = max_depth
        return root_node_id

    cdef Py_ssize_t new_leaf_node(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t parent,
        bint is_left,
    ) noexcept nogil:
        cdef Py_ssize_t node_id = self.tree.add_leaf_node(
            parent, is_left, end - start, self.criterion.weighted_n_total
        )
        self.criterion.leaf_value(self.tree, node_id)
        return node_id

    cdef Py_ssize_t new_branch_node(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        SplitPoint sp,
        Attribute *persistent,
        Py_ssize_t parent,
        bint is_left,
    ) noexcept nogil:
        cdef Py_ssize_t node_id
        node_id = self.tree.add_branch_node(
            parent,
            is_left,
            end - start,
            self.criterion.weighted_n_total,
            persistent,
            sp.threshold,
            sp.impurity_improvement,
        )
        return node_id

    cdef Py_ssize_t _build_tree(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t depth,
        Py_ssize_t parent,
        bint is_left,
        double impurity,
        Py_ssize_t *max_depth,
    ) noexcept nogil:
        """Recursive function for building the tree

        Each call to this function is allowed to access and transpose
        samples (in `self.samples`) in the region indicated by `start`
        and `end`.

        Parameters
        ----------
        start : int
            The start position in samples
        end : int
            The end position in samples
        depth : int (out)
            The current recusion depth
        parent : int
            The index of the parent node
        is_left : bool
            True if current node is left
        max_depth : int (out)
            The max depth reached in the tree
        """
        if depth > max_depth[0]:
            max_depth[0] = depth

        self.criterion.init(start, end, self.samples, self.sample_weights)
        cdef Py_ssize_t n_node_samples = end - start
        cdef bint is_leaf = (
            depth >= self.max_depth
            or n_node_samples < self.min_sample_split
            or n_node_samples < 2 * self.min_sample_leaf
        )
        if is_leaf:
            return self.new_leaf_node(start, end, parent, is_left)

        if parent < 0:
            impurity = self.criterion.impurity()

        cdef SplitPoint split = self._split(start, end, depth, impurity)

        cdef Attribute *persistent
        cdef Py_ssize_t current_node_id
        cdef Py_ssize_t err
        is_leaf = (
            split.split_point <= start
            or split.split_point >= end
            or split.impurity_improvement <= self.min_impurity_decrease
        )
        if not is_leaf:
            # The persistent attribute is freed by the Tree
            persistent = <Attribute*> malloc(sizeof(Attribute))
            err = self.generator.init_persistent(
                self.X, &split.attribute, persistent
            )
            self.generator.free_transient(&split.attribute)
            if err == -1:
                return -1

            current_node_id = self.new_branch_node(
                start, end, split, persistent, parent, is_left
            )
            self._build_tree(
                start,
                split.split_point,
                depth + 1,
                current_node_id,
                True,
                split.impurity_left,
                max_depth,
            )
            self._build_tree(
                split.split_point,
                end,
                depth + 1,
                current_node_id,
                False,
                split.impurity_right,
                max_depth,
            )
            return current_node_id
        else:
            return self.new_leaf_node(start, end, parent, is_left)

    cdef SplitPoint _split(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t depth,
        double parent_impurity
    ) noexcept nogil:
        cdef Py_ssize_t i, n_samples
        cdef Py_ssize_t current_split_point
        cdef double current_threshold
        cdef double current_impurity
        cdef double best_impurity
        cdef Attribute current_attribute

        n_samples = end - start

        best_impurity = -INFINITY

        current_attribute.attribute = NULL
        current_impurity = -INFINITY
        current_threshold = NAN
        current_split_point = 0

        cdef SplitPoint best
        best.threshold = NAN
        best.split_point = 0
        best.attribute.attribute = NULL

        for i in range(self.generator.get_n_attributess(self.X, depth)):
            self.generator.next_attribute(
                i,
                self.X,
                self.samples + start,
                n_samples,
                &current_attribute,
                &self.random_seed
            )

            self.generator.transient_values(
                &current_attribute,
                self.X,
                self.samples + start,
                end - start,
                self.attribute_buffer + start,
            )
            argsort(self.attribute_buffer + start, self.samples + start, n_samples)

            # All attribute values are constant
            if self.attribute_buffer[end - 1] <= self.attribute_buffer[start] + ATTRIBUTE_THRESHOLD:
                continue

            self.criterion.reset()
            self._partition_attribute_buffer(
                start, end, &current_split_point, &current_threshold, &current_impurity
            )
            if current_impurity > best_impurity:
                # store the order of samples in `sample_buffer`
                memcpy(
                    self.samples_buffer,
                    self.samples + start,
                    sizeof(Py_ssize_t) * n_samples,
                )
                best_impurity = current_impurity

                best.split_point = current_split_point
                best.threshold = current_threshold
                if best.attribute.attribute != NULL:
                    self.generator.free_transient(&best.attribute)

                best.attribute = current_attribute
            else:
                self.generator.free_transient(&current_attribute)

        # restore the best order to `samples`
        memcpy(
            self.samples + start,
            self.samples_buffer,
            sizeof(Py_ssize_t) * n_samples,
        )

        self.criterion.reset()
        self.criterion.update(start, best.split_point)
        self.criterion.child_impurity(&best.impurity_left, &best.impurity_right)
        best.impurity_improvement = self.criterion.impurity_improvement(
            parent_impurity,
            best.impurity_left,
            best.impurity_right,
            self.n_weighted_samples,
        )
        return best

    cdef void _partition_attribute_buffer(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t *best_split_point,
        double *best_threshold,
        double *best_impurity,
    ) noexcept nogil:
        cdef Py_ssize_t i  # real index of samples (in `range(start, end)`)
        cdef Py_ssize_t j  # sample index (in `samples`)
        cdef Py_ssize_t pos
        cdef Py_ssize_t new_pos
        cdef double impurity

        best_impurity[0] = -INFINITY
        best_threshold[0] = NAN
        best_split_point[0] = 0

        pos = start
        i = start
        while i < end:
            # Ignore split points with almost equal attribute value
            while i + 1 < end and (
                self.attribute_buffer[i + 1]
                <= self.attribute_buffer[i] + ATTRIBUTE_THRESHOLD
            ):
                i += 1

            i += 1
            if i < end:
                new_pos = i
                self.criterion.update(pos, new_pos)
                pos = new_pos
                impurity = self.criterion.proxy_impurity()
                if impurity > best_impurity[0]:
                    best_impurity[0] = impurity
                    best_threshold[0] = (
                        self.attribute_buffer[i - 1] / 2.0 + self.attribute_buffer[i] / 2.0
                    )
                    best_split_point[0] = pos

                    if (
                        best_threshold[0] == self.attribute_buffer[i]
                        or best_threshold[0] == INFINITY
                        or best_threshold[0] == -INFINITY
                    ):
                        best_threshold[0] = self.attribute_buffer[i - 1]

cdef class ExtraTreeBuilder(TreeBuilder):

    cdef void _partition_attribute_buffer(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t *split_point,
        double *threshold,
        double *impurity,
    ) noexcept nogil:
        cdef double min_attribute = self.attribute_buffer[start + 1]
        cdef double max_attribute = self.attribute_buffer[end - 1]
        cdef double rand_threshold = rand_uniform(
            min_attribute, max_attribute, &self.random_seed
        )
        cdef Py_ssize_t i

        split_point[0] = start
        for i in range(start + 1, end - 1):
            if self.attribute_buffer[i] <= rand_threshold:
                split_point[0] = i
            else:
                break
        threshold[0] = rand_threshold
        impurity[0] = INFINITY



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

from libc.math cimport NAN
from libc.stdlib cimport free, malloc
from numpy cimport uint32_t

from ..utils cimport TSArray


cdef class AttributeGenerator:

    cdef int reset(self, TSArray X) noexcept nogil:
        return 0

    cdef Py_ssize_t get_n_attributess(self, TSArray X) noexcept nogil:
        return -1

    cdef Py_ssize_t get_n_outputs(self, TSArray X) noexcept nogil:
        return self.get_n_attributess(X)

    cdef Py_ssize_t next_attribute(
        self,
        Py_ssize_t attribute_id,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        Attribute *transient,
        uint32_t *seed,
    ) noexcept nogil:
        return -1

    cdef Py_ssize_t free_transient(self, Attribute *attribute) noexcept nogil:
        return -1

    cdef Py_ssize_t free_persistent(self, Attribute *attribute) noexcept nogil:
        return -1

    cdef Py_ssize_t init_persistent(
        self,
        TSArray X,
        Attribute *transient,
        Attribute *persistent
    ) noexcept nogil:
        return 0

    cdef double transient_value(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample
    ) noexcept nogil:
        return NAN

    cdef double persistent_value(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample
    ) noexcept nogil:
        return NAN

    cdef Py_ssize_t transient_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t out_attribute,
    ) noexcept nogil:
        return -1

    cdef Py_ssize_t persistent_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t out_attribute,
    ) noexcept nogil:
        return -1

    cdef object persistent_to_object(self, Attribute *attribute):
        return None

    cdef Py_ssize_t persistent_from_object(self, object object, Attribute *attribute):
        return 0

    cdef void transient_values(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double* values
    ) noexcept nogil:
        cdef Py_ssize_t i
        for i in range(n_samples):
            values[i] = self.transient_value(attribute, X, samples[i])

    cdef void persistent_values(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double* values
    ) noexcept nogil:
        cdef Py_ssize_t i
        for i in range(n_samples):
            values[i] = self.persistent_value(attribute, X, samples[i])



"""
Fast distance computations.

The :py:mod:`wildboar.distance` module includes functions for computing
paired and pairwise distances between time series and between time series and
subsequences.

See the :ref:`User Guide <guide-metrics>` for more details and
examples.
"""

# Authors: Isak Samsten
# License: BSD 3 clause

from ._distance import (
    argmin_distance,
    argmin_subsequence_distance,
    check_metric,  # noqa: F401
    check_subsequence_metric,  # noqa: F401
    distance_profile,
    paired_distance,
    paired_subsequence_distance,
    paired_subsequence_match,
    pairwise_distance,
    pairwise_subsequence_distance,
    subsequence_match,
)
from ._manifold import MDS
from ._matrix_profile import matrix_profile
from ._neighbors import KMeans, KMedoids, KNeighborsClassifier

__all__ = [
    "argmin_distance",
    "argmin_subsequence_distance",
    "distance_profile",
    "pairwise_subsequence_distance",
    "paired_subsequence_distance",
    "subsequence_match",
    "paired_subsequence_match",
    "pairwise_distance",
    "paired_distance",
    "matrix_profile",
    "KNeighborsClassifier",
    "KMeans",
    "KMedoids",
    "MDS",
]



import numbers

from sklearn.metrics import silhouette_samples as sklearn_silhouette_samples
from sklearn.metrics import silhouette_score as sklearn_silhouette_score
from sklearn.utils._param_validation import (
    Interval,
    StrOptions,
    validate_params,
)
from sklearn.utils.validation import check_random_state

from ..distance._distance import _METRICS, pairwise_distance


@validate_params(
    {
        "X": ["array-like"],
        "labels": ["array-like"],
        "metric": [StrOptions(_METRICS.keys()), callable],
        "metric_params": [dict, None],
        "sample_size": [Interval(numbers.Integral, 1, None, closed="left"), None],
        "random_state": ["random_state"],
    },
    prefer_skip_nested_validation=True,
)
def silhouette_score(
    x,
    labels,
    *,
    metric="euclidean",
    metric_params=None,
    sample_size=None,
    random_state=None,
):
    """
    Compute the mean Silhouette Coefficient of all samples.

    Parameters
    ----------
    x : univariate time-series or multivariate time-series
        The input time series.
    labels : array-like of shape (n_samples,)
        Predicted labels for each sample.
    metric : str or callable, optional
        The metric to use when calculating distance between time series.
    metric_params : dict, optional
        The metric parameters. Read more about the metrics and their parameters
        in the :ref:`User guide <list_of_metrics>`.
    sample_size : int, optional
        The size of the sample to use when computing the Silhouette Coefficient
        on a random subset of the data.
        If ``sample_size is None``, no sampling is used.
    random_state : int or RandomState, optional
        Determines random number generation for selecting a subset of samples.
        Used when ``sample_size is not None``.

    Returns
    -------
    float
        Mean Silhouette Coefficient for all samples.

    Notes
    -----
    This is a convenient wrapper around :ref:`sklearn.metrics.silhouette_score`
    using Wildboar native metrics.
    """
    if sample_size is not None:
        random_state = check_random_state(random_state)
        idx = random_state.choice(x.shape[0], sample_size, replace=False)
        x = x[idx]
        labels = labels[idx]

    return sklearn_silhouette_score(
        pairwise_distance(x, metric=metric, metric_params=metric_params, dim="mean"),
        labels,
        metric="precomputed",
        sample_size=None,
        random_state=None,
    )


@validate_params(
    {
        "X": ["array-like"],
        "labels": ["array-like"],
        "metric": [StrOptions(_METRICS.keys()), callable],
        "metric_params": [dict, None],
    },
    prefer_skip_nested_validation=True,
)
def silhouette_samples(x, labels, *, metric="euclidean", metric_params=None):
    """
    Compute the Silhouette Coefficient of each samples.

    Parameters
    ----------
    x : univariate time-series or multivariate time-series
        The input time series.
    labels : array-like of shape (n_samples,)
        Predicted labels for each sample.
    metric : str or callable, optional
        The metric to use when calculating distance between time series.
    metric_params : dict, optional
        The metric parameters. Read more about the metrics and their parameters
        in the :ref:`User guide <list_of_metrics>`.

    Returns
    -------
    ndarray of shape (n_samples, )
        Silhouette Coefficient for each samples.

    Notes
    -----
    This is a convenient wrapper around :ref:`sklearn.metrics.silhouette_samples`
    using Wildboar native metrics.
    """
    return sklearn_silhouette_samples(
        pairwise_distance(x, metric=metric, metric_params=metric_params, dim="mean"),
        labels,
        metric="precomputed",
    )
